/*
 * data.c
 * code: SASANO Takayoshi
 * --- public domain, no warranty
 */

#include "common.h"

/* $B%G!<%?=hM}%?%9%/(B */
LOCAL	void	data_task(UW val)
{
	ERR	err;
	W	i, size;
	T_RTSK	rtsk;
	usbMsg	usbmsg;
	WorkArea	*wk;

	/* $B%?%9%/>pJs$N<hF@(B */
	err = ref_tsk(&rtsk, TSK_SELF);
	if (err < ER_OK) {
		D(0, ("data_task: ref_tsk %d\n", err));
		goto fin0;
	}
	wk = rtsk.exinf;

	while (1) {
		/* $BMW5aBT$A(B */
		err = rcv_mbf(&usbmsg, &size, wk->data_mbfid);
		if (err < ER_OK) goto fin1;

		/* $B%5%$%:$N0[$J$k%Q%1%C%H$OL5;k(B */
		if (size < sizeof(usbMsg)) {
			D(0, ("data_task: rcv_mbf size %d\n", size));
			continue;
		}

		/* $B<u?.$7$?%G!<%?$rI=<((B */
		printf("%d: ", usbmsg.pid);
		for (i = 0; i < usbmsg.datacnt; i++) {
			printf("%02x ", wk->data[i]);
		}
		printf("(%d)\n", usbmsg.error);

		/* STALL$B$J$i2sI|=hM}$r9T$&(B - $B%(%i!<%A%'%C%/$OITMW(B */
		if (usbmsg.error == USB_ERR_STALL) usbStallPipe(usbmsg.pid, 0);

		/* $B<!$N%G!<%?FI$_=P$7MW5a(B - $B%(%i!<%A%'%C%/$OITMW(B */
		usbIoPipe(usbmsg.pid, wk->data, wk->device_pktsz, NULL);
	}
 fin1:
	del_mbf(wk->data_mbfid);
 fin0:
	exd_tsk();
}

/* $B%G!<%?=hM}%?%9%/!&F0:n3+;O(B */
EXPORT	ERR	data_start(WorkArea *wk)
{
	WERR	err;
	T_CMBF	cmbf;
	T_CTSK	ctsk;

	/* $B%G!<%?<u?.MQ%a%C%;!<%8%P%C%U%!$N:n@.(B */
	cmbf.exinf  = wk;
	cmbf.mbfatr = TA_TFIFO;
	cmbf.bufsz  = sizeof(usbMsg);
	cmbf.maxmsz = sizeof(usbMsg);
	err = vcre_mbf(&cmbf);
	if (err < ER_OK) {
		D(0, ("data_start: vcre_mbf %d\n", err));
		goto fin0;
	}
	wk->data_mbfid = (ID)err;

	/* $B%G!<%?<u?.MQ%?%9%/$N:n@.(B */
	ctsk.exinf   = wk;
	ctsk.task    = data_task;
	ctsk.itskpri = DATA_PRI;
	ctsk.stksz   = DATA_STKSZ;
	ctsk.tskatr  = TA_HLNG | TA_RNG0;
	err = vcre_tsk(&ctsk);
	if (err < ER_OK) {
		D(0, ("data_start: vcre_tsk %d\n", err));
		goto fin1;
	}
	wk->data_tskid = (ID)err;

	/* $B%G!<%?<u?.MQ%?%9%/$N5/F0(B */
	err = sta_tsk(wk->data_tskid, 0);
	if (err < ER_OK) {
		D(0, ("data_start: sta_tsk %d\n", err));
		goto fin2;
	}

	err = ER_OK;
	goto fin0;

 fin2:
	del_tsk(wk->data_tskid);
 fin1:
	del_mbf(wk->data_mbfid);
 fin0:
	return err;
}

/* $B%G!<%?=hM}%?%9%/!&F0:n=*N;(B */
EXPORT	ERR	data_finish(WorkArea *wk)
{
	if (wk != NULL) {
		del_mbf(wk->data_mbfid);
	}

	return ER_OK;
}
