/*
 * device.c
 * code: SASANO Takayoshi
 * --- public domain, no warranty
 */

#include "common.h"

/* $BBP>]$H$9$k%G%P%$%9$+$I$&$+3NG'(B */
LOCAL	ERR	check_target(W did)
{
	ERR	err;
	usbDeviceDescriptor	d;

	err = usbDescriptorDevice(did, (UB *)&d, sizeof(d), NULL);
	if (err < 0) {
		D(0, ("check_target: usbDescriptorDevice %d\n", err));
		goto fin0;
	}

	/* idVendor, idProduct$B$N%A%'%C%/$r9T$&$,!":#2s$O2?$b$7$J$$(B */

 fin0:
	return err;
}

/* $B%G%P%$%9$N=i4|2=(B */
LOCAL	ERR	device_init(W did, W bnumif, W bepadr)
{
	ERR	err;
	usbDeviceRequest	r;

	/* SET_PROTOCOL(Report Protocol)$B$rH/9T$9$k(B */
	r.bmRequestType = 0x21;		/* bmR_CLASS | bmR_INTERFACE */
	r.bRequest	= 0x0b;		/* SET_PROTOCOL */
	r.wValue	= 0x0001;	/* Report Protocol */
	r.wIndex	= bnumif;

	err = usbRequestDevice(did, &r, NULL, 0, NULL);
	if (err < ER_OK) {
		D(0, ("device_init: usbRequestDevice %d\n", err));
		goto fin0;
	}

 fin0:
	return err;
}

/* endpoint descriptor$B$N8!:w(B */
LOCAL	VP	find_nextdesc(VP vp, W size, W *remain, W type)
{
	W	pos = 0;
	UB	*ptr = vp;

	while (pos < (size - 1)) {
		/* $BIT@5$J(Bdescriptor$B$r2sHr$9$k(B */
		if (ptr[pos] < 2) {
			D(0, ("find_nextdesc: ptr[%d] %d\n", pos, ptr[pos]));
			break;
		}

		/* $B<!$KMh$k(Bdescriptor$B$,(Btype$B$G;XDj$5$l$?J*$G$"$l$P=*N;(B */
		pos += ptr[pos];
		if (ptr[pos + 1] == type) {
			if (remain != NULL) *remain = size - pos;
			return &ptr[pos];
		}
	}

	return NULL;
}

/* endpoint address (interrupt, IN)$B$N8!:w(B */
LOCAL	ERR	find_endpoint(W iid, W *epadr, W *pktsz)
{
	ERR	err;
	UB	*buf;
	W	i, size, bnumep;
	usbEndpointDescriptor	*epdesc;

	/* interface descriptor$B$N%5%$%:$rF@$k(B */
	err = usbDescriptorInterface(iid, NULL, 0, &size);
	if (err < ER_OK) {
		D(0, ("find_endpoint: usbDescriptorInterface %d\n", err));
		goto fin0;
	}

	/* $B%a%b%j$N3NJ](B */
	buf = Kmalloc(size);
	if (buf == NULL) {
		D(0, ("find_endpoint: Kmalloc NULL\n"));
		err = ER_NOMEM;
		goto fin0;
	}

	/* interface descriptor$B$NFI$_=P$7(B */
	err = usbDescriptorInterface(iid, buf, size, NULL);
	if (err < ER_OK) {
		D(0, ("find_endpoint: usbDescriptorInterface %d\n", err));
		goto fin1;
	}

	/* interrupt IN pipe$B$N8!:w(B */
	epdesc = (usbEndpointDescriptor *)buf;
	bnumep = ((usbInterfaceDescriptor *)buf)->bNumEndpoints;

	err = ER_NOMDA;		/* bNumEndpoints = 0$B$H$$$&$3$H$bM-$jF@$k(B */
	for (i = 0; i < bnumep; i++) {
		epdesc = find_nextdesc(epdesc, size, &size, USB_ENDPOINT);

		/* endpoint descriptor$B$,8+$D$+$i$J$$(B */
		if (epdesc == NULL) break;

		/* interrupt/IN endpoint$B$J$i(Bendpoint address$B$r<hF@$7$F=*N;(B */
		if (epdesc->bmAttributes == USB_INTERRUPT &&
		    (epdesc->bEndpointAddress & 0x80)) {
			*epadr = epdesc->bEndpointAddress;
			*pktsz = epdesc->wMaxPacketSize;
			err = ER_OK;
			break;
		}
	}

	D(3, ("find_endpoint: %x\n", err));
 fin1:
	Kfree(buf);
 fin0:
	return err;
}

/* $B%G%P%$%9@\B3=hM}(B */
LOCAL	ERR	device_attach(WorkArea *wk, DevReq *q, DevRsp *p)
{
	WERR	err;
	W	did, iid, pid, bnumif, bepadr, wpktsz;

	D(3, ("device_attach: attach\n"));

	/* $B4{$K@\B3$5$l$F$$$k$J$i5qH](B */
	if (wk->device_did >= 0) {
		D(0, ("device_attach: already attached\n"));
		err = ER_BUSY;
		goto fin0;
	}

	/* device ID, bNumInterface$B$rF@$k(B */
	did    = q->devid & 0x7f;
	bnumif = (q->devid >> 8) & 0xff;

	/* $BBP>]$H$9$k%G%P%$%9$+$I$&$+%A%'%C%/(B */
	err = check_target(did);
	if (err < ER_OK) {
		D(0, ("device_attach: check_target %d\n", err));
		goto fin0;
	}

	/* $B%$%s%?%U%'!<%9$N%*!<%W%s(B */
	err = usbOpenInterface(did, bnumif);
	if (err < ER_OK) {
		D(0, ("device_attach: usbOpenInterface %d\n", err));
		goto fin0;
	}
	iid = err;

	/* $B%(%s%I%]%$%s%H%"%I%l%9$rF@$k(B */
	err = find_endpoint(iid, &bepadr, &wpktsz);
	if (err < ER_OK) {
		D(0, ("device_attach: find_endpoint %d\n", err));
		goto fin1;
	}

	/* $B%G%P%$%9$K8GM-$N@_Dj$r9T$&(B */
	err = device_init(did, bnumif, bepadr);
	if (err < ER_OK) {
		D(0, ("device_attach: device_init %d\n", err));
		goto fin1;
	}

	/* $B%Q%$%W$N%*!<%W%s(B */
	err = usbOpenPipe(iid, bepadr,
			  USB_NOWAIT | USB_SHORTOK, wk->data_mbfid);
	if (err < ER_OK) {
		D(0, ("device_attach: usbOpenPipe %d\n", err));
		goto fin1;
	}
	pid = err;

	/* $B:G=i$N%G!<%?FI$_=P$7MW5a$r=P$7$F$*$/(B */
	if (wpktsz > DATASIZE) wpktsz = DATASIZE;
	err = usbIoPipe(pid, wk->data, wpktsz, NULL);
	if (err < ER_OK) {
		D(0, ("device_attach: usbIoPipe %d\n", err));
		goto fin2;
	}

	/* $B%o!<%/%(%j%"$K3F<o(BID$B$rJ]B8$7$F=*N;(B */
	wk->device_did   = did;
	wk->device_iid   = iid;
	wk->device_pid   = pid;
	wk->device_pktsz = wpktsz;
	err = ER_OK;
	goto fin0;

 fin2:
	usbClosePipe(pid);
 fin1:
	usbCloseInterface(iid);
 fin0:
	return err;
}

/* $B%G%P%$%92r=|=hM}(B */
EXPORT	ERR	device_detach(WorkArea *wk, DevReq *q, DevRsp *p)
{
	ERR	err;

	/* $B8=:_;HMQCf$N%G%P%$%9$G$J$$>l9g$O5qH](B */
	if (q != NULL && (q->devid & 0x7f) != wk->device_did) {
		err = ER_NODEV;
		goto fin0;
	}

	usbClosePipe(wk->device_pid);
	usbCloseInterface(wk->device_iid);
	wk->device_did = wk->device_iid = wk->device_pid = -1;
	err = ER_OK;
 fin0:
	return err;
}

/* $B%G%P%$%9=hM}MW5a(B */
LOCAL	ERR	device_request(WorkArea *wk, DevReq *q, DevRsp *p)
{
	ERR	err;

	switch (q->cmd.cmd) {
	default:
		/* $BDL>oMW5aEy$OBP1~$7$J$$(B */
		err = ER_NOSPT;
		break;

	case	DC_ABORT:
	case	DC_SUSPEND:
	case	DC_RESUME:
	case	DC_CARDEVENT:
		/* $BL5;k(B */
		err = ER_OK;
		break;

	case	DC_USBEVENT:
		if (q->datano == USB_MOUNT) {
			err = (device_attach(wk, q, p) < ER_OK) ?
				USB_NONE : USB_OWN;
		} else {
			err = (device_detach(wk, q, p) < ER_OK) ?
				USB_NONE : USB_OWN;
		}
		break;
	}

	return err;
}

/* $B%G%P%$%9=hM}%?%9%/(B */
LOCAL	void	device_task(UW val)
{
	ERR	err;
	W	size;
	RNO	rno;
	DevReq	q;
	DevRsp	p;
	T_RTSK	rtsk;
	usbEventPattern	evp;
	WorkArea	*wk;

	/* $B%?%9%/>pJs$N<hF@(B */
	err = ref_tsk(&rtsk, TSK_SELF);
	if (err < ER_OK) {
		D(0, ("device_task: ref_tsk %d\n", err));
		goto fin0;
	}
	wk = rtsk.exinf;

	/* USB$B%$%s%?%U%'!<%9%I%i%$%P$H$7$FEPO?(B */
	evp.bClass    = 0x03;	/* HID Class */
	evp.bSubClass = 0x01;	/* Boot Interface Subclass */
	evp.bProtocol = 0x00;	/* not specified */
	evp.mask      = EVENT_CLASS | EVENT_SUBCLASS;
	err = usbRegistInterface(wk->device_porid, &evp);
	if (err < ER_OK) {
		D(0, ("device_task: usbRegistInterface %d\n", err));
		goto fin1;
	}

	while (1) {
		/* $BMW5aBT$A(B */
		err = acp_por(&rno, (VP)&q, &size,
			      wk->device_porid, D_ABORT_PTN);
		if (err < ER_OK) goto fin2;

		/* $B%5%$%:$N0[$J$k%Q%1%C%H$OL5;k(B */
		if (size < sizeof(q)) {
			D(0, ("device_task: rcp_por size %d\n", size));
			continue;
		}

		/* $BJVEzMQ%Q%1%C%H$N=`Hw$*$h$SMW5a=hM}$N<B9T(B */
		p.devid     = q.devid;
		p.cmd       = q.cmd;
		p.datano    = q.datano;
		p.datacnt   = 0;
		p.error.err = device_request(wk, &q, &p);

		/* $B%i%s%G%V$X$NJVEz(B */
		err = rpl_rdv(rno, (VP)&p, sizeof(p));
		if (err < 0) D(0, ("device_task: rpl_rdv %d\n", err));
	}
 fin2:
	usbRegistInterface(wk->device_porid, NULL);
 fin1:
	del_por(wk->device_porid);
 fin0:
	exd_tsk();
}

/* $B%G%P%$%9=hM}%?%9%/!&F0:n3+;O(B */
EXPORT	ERR	device_start(WorkArea *wk)
{
	WERR	err;
	T_CPOR	cpor;
	T_CTSK	ctsk;

	/* $B%i%s%G%V%]!<%H:n@.(B */
	cpor.exinf   = wk;
	cpor.poratr  = TA_NULL;
	cpor.maxcmsz = sizeof(DevReq);
	cpor.maxrmsz = sizeof(DevRsp);
	err = vcre_por(&cpor);
	if (err < ER_OK) {
		D(0, ("device_start: vcre_por %d\n", err));
		goto fin0;
	}
	wk->device_porid = (ID)err;

	/* $B%?%9%/:n@.(B */
	ctsk.exinf   = wk;
	ctsk.task    = device_task;
	ctsk.itskpri = DEVICE_PRI;
	ctsk.stksz   = DEVICE_STKSZ;
	ctsk.tskatr  = TA_HLNG | TA_RNG0;
	err = vcre_tsk(&ctsk);
	if (err < ER_OK) {
		D(0, ("device_start: vcre_tsk %d\n", err));
		goto fin1;
	}
	wk->device_tskid = (ID)err;

	/* $B%?%9%/5/F0(B */
	err = sta_tsk(wk->device_tskid, 0);
	if (err < ER_OK) {
		D(0, ("device_start: sta_tsk %d\n", err));
		goto fin2;
	}

	err = ER_OK;
	goto fin0;

 fin2:
	del_tsk(wk->device_tskid);
 fin1:
	del_por(wk->device_porid);
 fin0:
	return err;
}

/* $B%G%P%$%9=hM}%?%9%/!&F0:n=*N;(B */
EXPORT	ERR	device_finish(WorkArea *wk)
{
	if (wk != NULL) {
		del_por(wk->device_porid);
	}

	return ER_OK;
}
